package app

import (
	"context"
	"fmt"
	"os"

	"coopcloud.tech/abra/cli/internal"
	"coopcloud.tech/abra/pkg/autocomplete"
	"coopcloud.tech/abra/pkg/client"
	"coopcloud.tech/abra/pkg/log"
	stack "coopcloud.tech/abra/pkg/upstream/stack"
	"github.com/AlecAivazis/survey/v2"
	"github.com/docker/docker/api/types"
	"github.com/urfave/cli"
)

var appRemoveCommand = cli.Command{
	Name:      "remove",
	Aliases:   []string{"rm"},
	ArgsUsage: "<domain>",
	Usage:     "Remove all app data, locally and remotely",
	Description: `
This command removes everything related to an app which is already undeployed.

By default, it will prompt for confirmation before proceeding. All secrets,
volumes and the local app env file will be deleted.

Only run this command when you are sure you want to completely remove the app
and all associated app data. This is a destructive action, Be Careful!

If you would like to delete specific volumes or secrets, please use removal
sub-commands under "app volume" and "app secret" instead.

Please note, if you delete the local app env file without removing volumes and
secrets first, Abra will *not* be able to help you remove them afterwards.

To delete everything without prompt, use the "--force/-f" or the "--no-input/n"
flag.`,
	Flags: []cli.Flag{
		internal.ForceFlag,
		internal.DebugFlag,
		internal.NoInputFlag,
		internal.OfflineFlag,
	},
	BashComplete: autocomplete.AppNameComplete,
	Before:       internal.SubCommandBefore,
	Action: func(c *cli.Context) error {
		app := internal.ValidateApp(c)

		if !internal.Force && !internal.NoInput {
			log.Warnf("ALERTA ALERTA: this will completely remove %s data and config locally and remotely", app.Name)

			response := false
			prompt := &survey.Confirm{Message: "are you sure?"}
			if err := survey.AskOne(prompt, &response); err != nil {
				log.Fatal(err)
			}

			if !response {
				log.Fatal("aborting as requested")
			}
		}

		cl, err := client.New(app.Server)
		if err != nil {
			log.Fatal(err)
		}

		deployMeta, err := stack.IsDeployed(context.Background(), cl, app.StackName())
		if err != nil {
			log.Fatal(err)
		}
		if deployMeta.IsDeployed {
			log.Fatalf("%s is still deployed. Run \"abra app undeploy %s\"", app.Name, app.Name)
		}

		fs, err := app.Filters(false, false)
		if err != nil {
			log.Fatal(err)
		}

		secretList, err := cl.SecretList(context.Background(), types.SecretListOptions{Filters: fs})
		if err != nil {
			log.Fatal(err)
		}

		secrets := make(map[string]string)
		var secretNames []string

		for _, cont := range secretList {
			secrets[cont.Spec.Annotations.Name] = cont.ID // we have to map the names to ID's
			secretNames = append(secretNames, cont.Spec.Annotations.Name)
		}

		if len(secrets) > 0 {
			for _, name := range secretNames {
				err := cl.SecretRemove(context.Background(), secrets[name])
				if err != nil {
					log.Fatal(err)
				}
				log.Info(fmt.Sprintf("secret: %s removed", name))
			}
		} else {
			log.Info("no secrets to remove")
		}

		fs, err = app.Filters(false, true)
		if err != nil {
			log.Fatal(err)
		}

		volumeList, err := client.GetVolumes(cl, context.Background(), app.Server, fs)
		if err != nil {
			log.Fatal(err)
		}
		volumeNames := client.GetVolumeNames(volumeList)

		if len(volumeNames) > 0 {
			err := client.RemoveVolumes(cl, context.Background(), volumeNames, internal.Force, 5)
			if err != nil {
				log.Fatalf("removing volumes failed: %s", err)
			}

			log.Infof("%d volumes removed successfully", len(volumeNames))
		} else {
			log.Info("no volumes to remove")
		}

		if err = os.Remove(app.Path); err != nil {
			log.Fatal(err)
		}

		log.Info(fmt.Sprintf("file: %s removed", app.Path))

		return nil
	},
}
