package app

import (
	"context"
	"fmt"

	"coopcloud.tech/abra/cli/internal"
	"coopcloud.tech/abra/pkg/autocomplete"
	"coopcloud.tech/abra/pkg/client"
	"coopcloud.tech/abra/pkg/formatter"
	"coopcloud.tech/abra/pkg/log"
	"coopcloud.tech/abra/pkg/upstream/stack"
	"github.com/AlecAivazis/survey/v2"
	"github.com/urfave/cli"
)

var appVolumeListCommand = cli.Command{
	Name:      "list",
	Aliases:   []string{"ls"},
	ArgsUsage: "<domain>",
	Flags: []cli.Flag{
		internal.DebugFlag,
		internal.NoInputFlag,
	},
	Before:       internal.SubCommandBefore,
	Usage:        "List volumes associated with an app",
	BashComplete: autocomplete.AppNameComplete,
	Action: func(c *cli.Context) error {
		app := internal.ValidateApp(c)

		cl, err := client.New(app.Server)
		if err != nil {
			log.Fatal(err)
		}

		filters, err := app.Filters(false, true)
		if err != nil {
			log.Fatal(err)
		}

		volumes, err := client.GetVolumes(cl, context.Background(), app.Server, filters)
		if err != nil {
			log.Fatal(err)
		}

		headers := []string{"name", "created", "mounted"}

		table, err := formatter.CreateTable()
		if err != nil {
			log.Fatal(err)
		}

		table.Headers(headers...)

		var rows [][]string
		for _, volume := range volumes {
			row := []string{volume.Name, volume.CreatedAt, volume.Mountpoint}
			rows = append(rows, row)
		}

		table.Rows(rows...)

		if len(rows) > 0 {
			fmt.Println(table)
			return nil
		}

		log.Warnf("no volumes created for %s", app.Name)

		return nil
	},
}

var appVolumeRemoveCommand = cli.Command{
	Name:  "remove",
	Usage: "Remove volume(s) associated with an app",
	Description: `
This command supports removing volumes associated with an app. The app in
question must be undeployed before you try to remove volumes. See "abra app
undeploy <domain>" for more.

The command is interactive and will show a multiple select input which allows
you to make a seclection. Use the "?" key to see more help on navigating this
interface.

Passing "--force/-f" will select all volumes for removal. Be careful.`,
	ArgsUsage: "<domain>",
	Aliases:   []string{"rm"},
	Flags: []cli.Flag{
		internal.DebugFlag,
		internal.NoInputFlag,
		internal.ForceFlag,
	},
	Before:       internal.SubCommandBefore,
	BashComplete: autocomplete.AppNameComplete,
	Action: func(c *cli.Context) error {
		app := internal.ValidateApp(c)

		cl, err := client.New(app.Server)
		if err != nil {
			log.Fatal(err)
		}

		deployMeta, err := stack.IsDeployed(context.Background(), cl, app.StackName())
		if err != nil {
			log.Fatal(err)
		}

		if deployMeta.IsDeployed {
			log.Fatalf("%s is still deployed. Run \"abra app undeploy %s\"", app.Name, app.Name)
		}

		filters, err := app.Filters(false, true)
		if err != nil {
			log.Fatal(err)
		}

		volumeList, err := client.GetVolumes(cl, context.Background(), app.Server, filters)
		if err != nil {
			log.Fatal(err)
		}
		volumeNames := client.GetVolumeNames(volumeList)

		var volumesToRemove []string
		if !internal.Force && !internal.NoInput {
			volumesPrompt := &survey.MultiSelect{
				Message: "which volumes do you want to remove?",
				Help:    "'x' indicates selected, enter / return to confirm, ctrl-c to exit, vim mode is enabled",
				VimMode: true,
				Options: volumeNames,
				Default: volumeNames,
			}
			if err := survey.AskOne(volumesPrompt, &volumesToRemove); err != nil {
				log.Fatal(err)
			}
		}

		if internal.Force || internal.NoInput {
			volumesToRemove = volumeNames
		}

		if len(volumesToRemove) > 0 {
			err := client.RemoveVolumes(cl, context.Background(), volumesToRemove, internal.Force, 5)
			if err != nil {
				log.Fatalf("removing volumes failed: %s", err)
			}

			log.Infof("%d volumes removed successfully", len(volumesToRemove))
		} else {
			log.Info("no volumes removed")
		}

		return nil
	},
}

var appVolumeCommand = cli.Command{
	Name:      "volume",
	Aliases:   []string{"vl"},
	Usage:     "Manage app volumes",
	ArgsUsage: "<domain>",
	Subcommands: []cli.Command{
		appVolumeListCommand,
		appVolumeRemoveCommand,
	},
}
